from abc import ABC, abstractmethod
import logging
from typing import List, Dict, Any, Union

from siglab_py.algo.strategy_base import StrategyBase

logger = logging.getLogger()

class MACDRSICrosses15mTCStrategy(StrategyBase):
    def __init__(self, *args: object) -> None:
        pass
    
    @staticmethod
    def order_notional_adj(
        algo_param : Dict,
    ) -> Dict[str, float]:
        return StrategyBase.order_notional_adj(algo_param) # type: ignore

    @staticmethod
    def allow_entry_initial(
        lo_row_tm1,
        hi_row_tm1,
        last_candles
    )  -> Dict[str, bool]:
        logger.info(f"********allow_entry_initial ********")
        logger.info(f"lo_row_tm1['datetime']: {lo_row_tm1['datetime']}, hi_row_tm1['datetime']: {hi_row_tm1['datetime']}")
        logger.info(f"macd_cross: {lo_row_tm1['macd_cross']}, lo_row_tm1['rsi_trend']: {lo_row_tm1['rsi_trend']}")
        logger.info(f"lo_row_tm1['close']: {lo_row_tm1['close']}, hi_row_tm1['ema_close']: {hi_row_tm1['ema_close']}")

        allow_long : bool = (
                lo_row_tm1['macd_cross'] == 'bullish' 
                # use 'macd_cross_last' instead in combinations with 'macd_bullish_cross_last_id' if you want to make more entries
                '''
                and (
                      lo_row_tm1.name >= lo_row_tm1['macd_bullish_cross_last_id']
                      and 
                      (lo_row_tm1.name - lo_row_tm1['macd_bullish_cross_last_id']) < 5
                )
                '''
                and lo_row_tm1['rsi_trend']=="up"
                and lo_row_tm1['close']>hi_row_tm1['ema_close']
        )
        allow_short : bool = (
                lo_row_tm1['macd_cross'] == 'bearish'
                '''
                and (
                      lo_row_tm1.name >= lo_row_tm1['macd_bearish_cross_last_id']
                      and 
                      (lo_row_tm1.name - lo_row_tm1['macd_bearish_cross_last_id']) < 5
                )
                '''
                and lo_row_tm1['rsi_trend']=="down"
                and lo_row_tm1['close']<hi_row_tm1['ema_close']
        )
        return {
            'long' : allow_long,
            'short' : allow_short
        }

    @staticmethod
    def allow_entry_final(
        lo_row,
        algo_param : Dict
    ) -> Dict[str, Union[bool, float, None]]:
        open : float = lo_row['open']

        entry_price_long, entry_price_short = open, open
        allow_long, allow_short = True, True
        reference_price = None
        
        pnl_potential_bps = algo_param['tp_max_percent']*100

        target_price_long = entry_price_long * (1 + pnl_potential_bps/10000)
        target_price_short = entry_price_short * (1 - pnl_potential_bps/10000)

        return {
                'long' : allow_long,
                'short' : allow_short,

                # In additional to allow or not, allow_entry_final also calculate a few things which you may need to mark the entry trades.
                'entry_price_long' : entry_price_long,
                'entry_price_short' : entry_price_short,
                'target_price_long' : target_price_long,
                'target_price_short' : target_price_short,
                'reference_price' : reference_price
            }
    
    @staticmethod
    def trailing_stop_threshold_eval(
        algo_param : Dict,
        *args: Any, **kwargs: Any
    ) -> Dict[str, float]:
        result = StrategyBase.trailing_stop_threshold_eval(algo_param)
        tp_min_percent = result['tp_min_percent']
        tp_max_percent = result['tp_max_percent']
        return {
            'tp_min_percent' : tp_min_percent,
            'tp_max_percent' : tp_max_percent
        }

    @staticmethod
    def tp_eval (
        mid : float,
        tp_max_target : float,
        pos_side : OrderSide
    ) -> bool:
        return StrategyBase.tp_eval(mid, tp_max_target, pos_side)

    @staticmethod
    def get_strategy_indicators() -> List[str]:
        return [ 
            'lo_row_tm1:macd_cross', 'lo_row_tm1:macd_bullish_cross_last_id', 'lo_row_tm1:macd_bearish_cross_last_id', 
            'lo_row_tm1:rsi_trend',  'lo_row_tm1:rsi', 'lo_row_tm1:rsi_max', 'lo_row_tm1:rsi_min', 'lo_row_tm1:rsi_idmax', 'lo_row_tm1:rsi_idmin' 
        ]


from typing import Dict

import ccxt

if __name__ == '__main__':
    apiKey : str = None
    secret : str = None 
    passphrase : str = None
    subaccount : str = None
    
    params : Dict = {
        'apiKey' : apiKey,
        'secret' : secret,
        'password' : passphrase,
        'subaccount' : subaccount,
        'rateLimit' : 100,                    # In ms
        'options' : {
            'defaultType': 'swap', # 'funding', 'spot', 'margin', 'future', 'swap', 'option'
        }
    }
    exchange = ccxt.okx(params)
    markets = exchange.load_markets()

    if apiKey:
        balances = exchange.fetch_balance()
    
    xaut_perp_ticker = 'XAU/USDT:USDT'
    found = next(( market for market in markets if market == xaut_perp_ticker), None)
    if found:
        '''
        'id' = 'XAU-USDT-SWAP'
        'symbol' = 'XAU/USDT:USDT' <-- normalized symbol
        '''
        xaut_perp_market = markets[xaut_perp_ticker]
        print(xaut_perp_market)

        entry_order = exchange.create_order(
            symbol = xaut_perp_ticker,
            amount = 10, # buy 0.001 XAU means 10 XAUT/USDT:USDT contract (0.0001 multipler)
            type='market',
            side='sell'
        )

        print(entry_order)

        positions = exchange.fetch_positions()
        for pos in positions:
            '''
            For longs, 'side' = 'long', 'contracts' = 100.0 (positive integer)
            For shorts, 'side' = 'short', 'contracts' = 100.0 (positive integer)
            '''
            print(pos)

            exit_order = exchange.create_order(
                symbol = xaut_perp_ticker,
                amount = pos['contracts'],
                type='market',
                side='buy' if pos['side']=='short' else 'sell'
            )


'''
https://medium.com/@natalia_assad/how-send-a-table-to-slack-using-python-d1a20b08abe0
'''
import sys
from typing import Any, Dict
import json
import requests

from siglab_py.constants import LogLevel

def slack_dispatch_notification(
        title : str,
        message : str,
        footer : str,
        params : Dict[str, Any],
        log_level : LogLevel = LogLevel.INFO,
        max_message_len : int = 1800
):
    slack_params = params['slack']

    # Slack slack ... https://stackoverflow.com/questions/60344831/slack-api-invalid-block
    message = message[:max_message_len]

    if log_level.value==LogLevel.INFO.value or log_level.value==LogLevel.DEBUG.value:
        webhook_url = slack_params['info']['webhook_url']
    elif log_level.value==LogLevel.CRITICAL.value:
        webhook_url = slack_params['critical']['webhook_url']
    elif log_level.value==LogLevel.ERROR.value:
        webhook_url = slack_params['alert']['webhook_url']
    else:
        webhook_url = slack_params['info']['webhook_url']

    if not webhook_url:
        return

    data = {
        "username": "siglab_py",
        "type": "section",
        "blocks": [
            {
                "type": "header",
                "text": { "type": "plain_text", "text": f"{title}" }
            },
            {
                "type": "section",
                "text": { "type": "mrkdwn", "text": message }
            },
            {
                "type": "section",
                "text": { "type": "plain_text", "text": footer }
            }
        ]
    }

    byte_size = str(sys.getsizeof(data, 2000))
    req_headers = { 'Content-Length': byte_size, 'Content-Type': "application/json"}
    rsp = requests.post(webhook_url, headers=req_headers, data=json.dumps(data))
    if rsp.status_code != 200:
        raise Exception(rsp.status_code, rsp.text)


import sys
import traceback
from enum import Enum
import argparse
import time
from datetime import datetime, timedelta
from typing import Any, Dict, Union
import logging
import json
from tabulate import tabulate
import asyncio
from threading import Thread
from collections import defaultdict
import pandas as pd
import numpy as np
from redis import StrictRedis
from redis.client import PubSub

from ccxt.binance import binance
from ccxt.okx import okx
from ccxt.bybit import bybit
from ccxt.base.exchange import Exchange

from siglab_py.util.market_data_util import fetch_candles, get_old_ticker, get_ticker_map


'''
To start from command prompt:
    set PYTHONPATH=%PYTHONPATH%;D:\dev\siglab\siglab_py
    python candles_provider.py --provider_id aaa --candle_size 1h --how_many_candles 2169 --redis_ttl_ms 3600000

This script is pypy compatible:
    pypy candles_provider.py --provider_id aaa --candle_size 1h --how_many_candles 2169 --redis_ttl_ms 3600000

Key parameters you may want to modify:
    provider_id: You can trigger this provider instance using trigger_provider.py. Of course, you'd write your own.
    candle_size: 1m, 5m, 15min, 1h, 1d for example.
    how_many_candles: default to 2169 (24 x 90). 
    redis_ttl_ms: This is how long orderbook snapshot will last on redis when provider publishes to it.

Launch.json if you wish to debug from VSCode:
    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python Debugger: Current File",
                "type": "debugpy",
                "request": "launch",
                "program": "${file}",
                "console": "integratedTerminal",
                "args" : [
                            "--provider_id", "YourProviderNameHere",
                            "--candle_size", "1h",
                            "--how_many_candles", "2169",
                            "--redis_ttl_ms", "3600000"
                ],
                "env": {
                    "PYTHONPATH": "${workspaceFolder}"
                }
            }
        ]
    }
'''
class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

param : Dict = {
    "candle_size" : "1h",
    'how_many_candles' : 24*90,

    'market_type' : 'linear', # For spots, set to "spot". For perpectual, you need to look at ccxt doc, for most exchanges, it's 'linear' or 'swap' for perpetuals. Example, https://github.com/ccxt/ccxt/blob/master/python/ccxt/okx.py?plain=1#L1110
    
    # Provider ID is part of mds publish topic. 
    'provider_id' : 'b0f1b878-c281-43d7-870a-0347f90e6ece',

    # Ticker change
    'ticker_change_map' : '.\\siglab_py\\ticker_change_map.json',

    # Publish to message bus
    'mds' : {
        'topics' : {
            'partition_assign_topic' : 'mds_assign_$PROVIDER_ID$',
            'candles_publish_topic' : 'candles-$DENORMALIZED_SYMBOL$-$EXCHANGE_NAME$-$INTERVAL$'
        },
        'redis' : {
            'host' : 'localhost',
            'port' : 6379,
            'db' : 0,
            'ttl_ms' : 1000*60*15 # 15 min?
        }

    }    
}

logging.Formatter.converter = time.gmtime
logger = logging.getLogger()
log_level = logging.INFO # DEBUG --> INFO --> WARNING --> ERROR
logger.setLevel(log_level)
format_str = '%(asctime)s %(message)s'
formatter = logging.Formatter(format_str)
sh = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)
# fh = logging.FileHandler(f"{param['job_name']}.log")
# fh.setLevel(log_level)
# fh.setFormatter(formatter)     
# logger.addHandler(fh)

market_type : str = param['market_type'] 

binance_exchange = binance({
    'defaultType' : market_type
})

okx_exchange = okx({
    'defaultType' : market_type
})

bybit_exchange = bybit({
    'defaultType' : market_type
})

exchanges = {
    f"binance" : binance_exchange,
    f"okx" : okx_exchange,
    f"bybit" : bybit_exchange
}

def log(message : str, log_level : LogLevel = LogLevel.INFO):
    if log_level.value<LogLevel.WARNING.value:
        logger.info(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.WARNING.value:
        logger.warning(f"{datetime.now()} {message}")

    elif log_level.value==LogLevel.ERROR.value:
        logger.error(f"{datetime.now()} {message}")

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore

    parser.add_argument("--provider_id", help="candle_provider will go to work if from redis a matching topic partition_assign_topic with provider_id in it.", default=None)
    parser.add_argument("--candle_size", help="candle interval: 1m, 1h, 1d... etc", default='1h')
    parser.add_argument("--how_many_candles", help="how_many_candles", default=24*7)

    parser.add_argument("--redis_ttl_ms", help="TTL for items published to redis. Default: 1000*60*60 (i.e. 1hr)",default=1000*60*60)

    args = parser.parse_args()
    if args.provider_id:
        param['provider_id'] = args.provider_id
    param['candle_size'] = args.candle_size
    param['how_many_candles'] = int(args.how_many_candles)

    param['redis_ttl_ms'] = int(args.redis_ttl_ms)

def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = param['mds']['redis']['host'],
                    port = param['mds']['redis']['port'],
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
        raise ConnectionError(err_msg)
    
    return redis_client
    
def init_redis_channel_subscription(redis_client : StrictRedis, partition_assign_topic : str) -> PubSub:
    redis_client = init_redis_client()
    pubsub = redis_client.pubsub()
    pubsub.subscribe(partition_assign_topic)
    return pubsub

def process_universe(
    universe : pd.DataFrame, 
    ticker_change_map,
    task,
    redis_client : StrictRedis
    ):
    # Key = ticker
    subscribed : Dict[tuple, Dict[str, Any]]  = defaultdict(lambda : {'candles': None, 'num_candles': 0} )

    while task.keep_running:
        start = time.time()
        num_fetches_this_wave = 0

        i = 1
        for index, row in universe.iterrows():
            exchange_name : str = row['exchange']
            ticker : str = row['ticker']
            
            try:
                _ticker : str = ticker # In case ticker change
                if ticker_change_map:
                    old_ticker= get_old_ticker(_ticker, ticker_change_map)
                    if old_ticker:
                        ticker_change_mapping = get_ticker_map(reference_ticker, ticker_change_map)
                        ticker_change_cutoff_sec = int(ticker_change_mapping['cutoff_ms']) / 1000
                        if datetime.now().timestamp()<ticker_change_cutoff_sec:
                            _ticker = old_ticker
                
                this_row_header = f'({i} of {universe.shape[0]}) exchange_name: {exchange_name}, ticker: {_ticker}'
                
                exchange = exchanges[exchange_name]

                exchange.load_markets() # in case ticker change after gateway startup

                fetch_again = False
                last_fetch = None
                last_fetch_ts = None
                if subscribed[(exchange_name, _ticker)]:
                    last_fetch = subscribed[(exchange_name, _ticker)]['candles']
                    if subscribed[(exchange_name, _ticker)]['num_candles']>0:
                        last_fetch_ts = last_fetch.iloc[-1]['timestamp_ms']/1000 # type: ignore Otherwise, Error: Cannot access attribute "iloc" for class "None"
                candle_size = param['candle_size']
                interval = candle_size[-1]
                num_intervals_per_candle = int(candle_size.replace(interval,""))
                number_intervals = param['how_many_candles']
                
                start_date : datetime = datetime.now()
                end_date : datetime = start_date
                if interval=="m":
                    end_date = datetime.now()
                    end_date = datetime(end_date.year, end_date.month, end_date.day, end_date.hour, end_date.minute, 0)
                    start_date = end_date + timedelta(minutes=-num_intervals_per_candle*number_intervals) 

                    num_sec_since_last_fetch = (end_date.timestamp() - last_fetch_ts) if last_fetch_ts else sys.maxsize
                    fetch_again = True if num_sec_since_last_fetch >= 60 / 10 else False

                elif interval=="h":
                    end_date = datetime.now()
                    end_date = datetime(end_date.year, end_date.month, end_date.day, end_date.hour, 0, 0)
                    start_date = end_date + timedelta(hours=-num_intervals_per_candle*number_intervals) 

                    num_sec_since_last_fetch = (end_date.timestamp() - last_fetch_ts) if last_fetch_ts else sys.maxsize
                    fetch_again = True if num_sec_since_last_fetch >= 60*60 / 10 else False

                elif interval=="d":
                    end_date = datetime.now()
                    end_date = datetime(end_date.year, end_date.month, end_date.day, 0, 0, 0)
                    start_date = end_date + timedelta(days=-num_intervals_per_candle*number_intervals) 

                    num_sec_since_last_fetch = (end_date.timestamp() - last_fetch_ts) if last_fetch_ts else sys.maxsize
                    fetch_again = True if num_sec_since_last_fetch >= 24*60*60 / 10 else False
                
                cutoff_ts = int(start_date.timestamp()) # in seconds
                
                if fetch_again:
                    if datetime.now().minute==0:
                        time.sleep(10) # Give some time for the exchange

                    candles = fetch_candles(
                                                start_ts=cutoff_ts, 
                                                end_ts=int(end_date.timestamp()), 
                                                exchange=exchange, normalized_symbols=[_ticker], 
                                                candle_size = candle_size, 
                                                num_candles_limit = 100,
                                                logger = None
                                            )
                    subscribed[(exchange_name, _ticker)] = {
                        'candles' : candles[ _ticker ],
                        'num_candles' : candles[ _ticker ].shape[0] # type: ignore Otherwise, Error: "shape" is not a known attribute of "None"
                    }
                    num_fetches_this_wave += 1

                    denormalized_ticker = next(iter([ exchange.markets[x] for x in exchange.markets if exchange.markets[x]['symbol']==_ticker]))['id']

                    publish_key = param['mds']['topics']['candles_publish_topic']
                    publish_key = publish_key.replace('$DENORMALIZED_SYMBOL$', denormalized_ticker)
                    publish_key = publish_key.replace('$EXCHANGE_NAME$', exchange_name)
                    publish_key = publish_key.replace('$INTERVAL$', param['candle_size'])

                    data = candles[_ticker].to_json(orient='records') # type: ignore Otherwise, Error: "to_json" is not a known attribute of "None"
                    
                    start = time.time()
                    if redis_client:
                        '''
                        https://redis.io/commands/set/
                        '''
                        expiry_sec : int = 0
                        if interval=="m":
                            expiry_sec = 60 + 60*15
                        elif interval=="h":
                            expiry_sec = 60*60 + 60*15
                        elif interval=="d":
                            expiry_sec = 60*60*24 
                        expiry_sec += 60*15 # additional 15min

                        redis_client.set(name=publish_key, value=json.dumps(data).encode('utf-8'), ex=expiry_sec)

                        redis_set_elapsed_ms = int((time.time() - start) *1000)

                        log(f"published candles {candles[_ticker].shape[0]} rows. {this_row_header} {publish_key} {sys.getsizeof(data, -1)} bytes to mds elapsed {redis_set_elapsed_ms} ms")

            except Exception as loop_error:
                log(f"Failed to process {this_row_header}. Error: {loop_error} {str(sys.exc_info()[0])} {str(sys.exc_info()[1])} {traceback.format_exc()}")

            if not task.keep_running:
                break

            i += 1

        if num_fetches_this_wave>0:
            log(f"Fetch candles for whole universe done.  elapsed: {time.time()-start} sec, universe_reload_id: {task.universe_reload_id}. # tickers: {len(subscribed)}")
        else:
            log(f"universe_reload_id: {task.universe_reload_id}, Nothing to fetch this wave. Sleep a bit.")
            time.sleep(3)
    
    log(f"process_universe exit, universe_reload_id: {task.universe_reload_id}")
    
async def main():
    parse_args()
    
    param['job_name'] = f'candles_provider_{param["provider_id"]}'

    ticker_change_map = None
    with open(param['ticker_change_map'], 'r', encoding='utf-8') as f:
        ticker_change_map : List[Dict[str, Union[str, int]]] = json.load(f)
        log(f"ticker_change_map loaded from {param['ticker_change_map']}")
        log(json.dumps(ticker_change_map))

    redis_client : StrictRedis = init_redis_client()
    partition_assign_topic : str = param['mds']['topics']['partition_assign_topic']
    partition_assign_topic = partition_assign_topic.replace("$PROVIDER_ID$", param['provider_id'])
    redis_pubsub : PubSub = init_redis_channel_subscription(redis_client, partition_assign_topic)

    class ThreadTask:
        def __init__(self, universe_reload_id) -> None:
            self.keep_running = True
            self.universe_reload_id = universe_reload_id
    task = None

    log(f"candles_provider {param['provider_id']} started, waiting for trigger. (Can use trigger_provider.py to trigger it)")

    universe_reload_id = 1
    for message in redis_pubsub.listen():
        if message['type'] == 'message' and message['channel'].decode()==partition_assign_topic:
            if task:
                task.keep_running = False

            tickers = json.loads(message['data'].decode('utf-8'))
            tickers = [ { 'exchange' : x.split('|')[0], 'ticker' : x.split('|')[-1] } for x in tickers ]
            universe = pd.DataFrame(tickers)
            logger.info(f"{partition_assign_topic} {message}")

            task = ThreadTask(universe_reload_id=universe_reload_id)
            t = Thread(target=process_universe, args = (universe, ticker_change_map, task, redis_client))
            t.start()

            universe_reload_id += 1

async def _run_jobs():
    await main()
asyncio.get_event_loop().run_until_complete(_run_jobs())